//
// EGLE 0.5		10/14/97
// Quake map viewer using OpenGl API
// By Phil Frisbie, Jr.
// pfrisbie@geocities.com
//
// Inspired by QMAP (copyright 1997 Sean Barrett)
//
// Main OpenGL rendering code based on the SGI source code from
// 'Quake walkthrough in OpenGL' by Edward Hutchins
//
// And of course the QuakeEd source by id Software
//

#include "egle.h"

  
//
// globals
//

int				W = 640, H = 480;
//int				W = 180, H = 120; // this is good for software only
int				bVerbose = FALSE;
int				nPixelFormat = GL_RGBA;
int				nTexelFormat = 4;
double			dRealTime;
double			dSimTime;
unsigned int	uFrame = 0;
double			fps;
unsigned int	uElapsedFrames;
char			*pszMap = "start.bsp";
char			*pGLVendor;
char			*pGLRenderer;

PEPISODE		pCurrEpsd = NULL;

OBJECT			Player;

// rendering options
int				bDrawHulls = FALSE;
int				bDrawTexture = TRUE;
int				usHull = 0;
// I suspect 1 is the shot clipping hull, 2 is the walking clipping hull...
int				usClipHull = 2;

void SetViews( void )
{

	pCurrEpsd->pView[0] = malloc(sizeof(VIEW));

	pCurrEpsd->pView[0]->ScrX = 0;
	pCurrEpsd->pView[0]->ScrY = 0;
	pCurrEpsd->pView[0]->Width = W;
	pCurrEpsd->pView[0]->Height = H;
	pCurrEpsd->pView[0]->pObject = &Player;
	pCurrEpsd->pView[0]->ObjHeight = 0;
	pCurrEpsd->pView[0]->HAngle = (float)0;

	// enable this for the sample view-in-a-view
#if 1
	pCurrEpsd->pView[1] = malloc(sizeof(VIEW));
	pCurrEpsd->pView[1]->ScrX = 10;
	pCurrEpsd->pView[1]->ScrY = H - 10 - 128;
	pCurrEpsd->pView[1]->Width = 128;
	pCurrEpsd->pView[1]->Height = 128;
	pCurrEpsd->pView[1]->pObject = &Player;
	pCurrEpsd->pView[1]->ObjHeight = 0;
	pCurrEpsd->pView[1]->HAngle = (float)180;
#else
	pCurrEpsd->pView[1] = NULL;
#endif
	pCurrEpsd->pView[2] = NULL;
	pCurrEpsd->pView[3] = NULL;
}

PEPISODE CreateEpisode( void )
{
    PEPISODE	pEpsd = NULL;
    int			bSuccess = FALSE;

    pEpsd = malloc( sizeof(*pEpsd) );
    if (!pEpsd) goto Abort;

    // init the scratch texgen matrix
    pEpsd->M[ 2] = 0; pEpsd->M[ 3] = 0;
    pEpsd->M[ 6] = 0; pEpsd->M[ 7] = 0;
    pEpsd->M[10] = 0; pEpsd->M[11] = 0;
    pEpsd->M[14] = 0; pEpsd->M[15] = 1;

    bSuccess = TRUE;

Abort:
    if (!bSuccess) pEpsd = NULL;
    return( pEpsd );
}

int DestroyEpisode( PEPISODE pEpsd )
{
    if (pEpsd)
		free(pEpsd);
    return( TRUE );
}

void TessBegin( GLenum mode )
{
	if (pCurrEpsd->uNextSPA < pCurrEpsd->uSPAMax)
	{
     pCurrEpsd->uThisPrim = pCurrEpsd->uNextSPA++;
	    ++(pCurrEpsd->pSDL[pCurrEpsd->usTessSurf].uCount);
	    pCurrEpsd->pSPA[pCurrEpsd->uThisPrim].p.sMode = (short int)mode;
	    pCurrEpsd->pSPA[pCurrEpsd->uThisPrim].p.sCount = 0;
	}
	else if (pCurrEpsd->uNextSPA == pCurrEpsd->uSPAMax)
	{
	    printf( "ERROR: surface primitive array overflow!" );
	    pCurrEpsd->uNextSPA = pCurrEpsd->uSPAMax + 1;
	}
    
	glBegin( mode );
}

void TessVertex3dv( GLdouble *v )
{
	if (pCurrEpsd->uNextSPA < pCurrEpsd->uSPAMax)
	{
	    float * pf = pCurrEpsd->pSPA[(pCurrEpsd->uNextSPA)++].c.v;
	    *pf++ = (float)*v++;
	    *pf++ = (float)*v++;
	    *pf++ = (float)*v++;
	    ++(pCurrEpsd->pSPA[pCurrEpsd->uThisPrim].p.sCount);
	}
	else if (pCurrEpsd->uNextSPA == pCurrEpsd->uSPAMax)
	{
		printf( "ERROR: surface primitive array overflow!" );
		pCurrEpsd->uNextSPA = pCurrEpsd->uSPAMax + 1;
	}
}

void TessEnd( void )
{
	if (pCurrEpsd->pSDL[pCurrEpsd->usTessSurf].uCount < 1)
		printf( "ERROR: empty primitive!" );
    
	glEnd();
}

int TesselateEpisode( void )
{
    int				i, nColinElim = 0, nTotVerts;
    int				*pEL = NULL;
    int				*pContourEL = NULL;
    int				*pnContourSize = NULL;
    double			*pdVtx = NULL;
    surface_t		*pS = pCurrEpsd->pS;
    edge_t			*pE = NULL;
    GLUtesselator	*tess = NULL;
    GLfloat			M[16];
    int				bSuccess = FALSE;

    if (!pCurrEpsd) goto Abort;

    tess = gluNewTess();
    if (!tess) BAIL(( "Failed to create tesselator!" ));
    gluTessCallback( tess, GLU_TESS_BEGIN_DATA, (void *)TessBegin );
    gluTessCallback( tess, GLU_TESS_VERTEX_DATA, (void *)TessVertex3dv );
    gluTessCallback( tess, GLU_TESS_END_DATA, (void *)TessEnd );
    gluTessCallback( tess, GLU_TESS_COMBINE, NULL );
    gluTessProperty( tess, GLU_TESS_BOUNDARY_ONLY, GL_FALSE );

    pEL = pCurrEpsd->pEL;
    pE = pCurrEpsd->pE;

    // estimate the number of verticies needed
    for (i = nTotVerts = 0; i < pCurrEpsd->nS; ++i)
		nTotVerts += pS[i].numedge;

    // pad vertex count by 1/8th, allocate storage for vertex arrays
    pCurrEpsd->uSPAMax = nTotVerts + (nTotVerts >> 3);
    pCurrEpsd->uNextSPA = pCurrEpsd->uThisPrim = 0;
    Progress(( "%d total verticies in lists (padded to %d), %d in bsp\n", nTotVerts, pCurrEpsd->uSPAMax, pCurrEpsd->nV ));

    pCurrEpsd->pSDL = malloc( pCurrEpsd->nS * sizeof(pCurrEpsd->pSDL[0]) );
    pCurrEpsd->pSPA = malloc( pCurrEpsd->uSPAMax * sizeof(pCurrEpsd->pSPA[0]) );
    pCurrEpsd->pfTGM = malloc( pCurrEpsd->nTG * 8 * sizeof(pCurrEpsd->pfTGM[0]) );
    pContourEL = malloc( pCurrEpsd->nMaxEdgesPerSurface * sizeof(pContourEL[0]) );
    pnContourSize = malloc( pCurrEpsd->nMaxEdgesPerSurface * sizeof(pnContourSize[0]) );
    pdVtx = malloc( 2 * pCurrEpsd->nMaxEdgesPerSurface * 3 * sizeof(pdVtx[0]) );
    if (!pCurrEpsd->pSDL || !pCurrEpsd->pSPA || !pCurrEpsd->pfTGM || !pContourEL || !pnContourSize || !pdVtx)
		BAIL(( "Out of memory!" ));

    for (i = 0; i < pCurrEpsd->nS; ++i)
    {
		int		    j, k, nNumEdges = pS[i].numedge, nContours = 0;
		texgen_t	    *pTG = &pCurrEpsd->pTG[pS[i].texgenid];
		miptex_t	    *pMipTex = &pCurrEpsd->pMipTex[pTG->texid];
		// copy the edge array
		for (j = 0; j < nNumEdges; ++j)
		    pContourEL[j] = pEL[pS[i].firstedge + j];

		// topo-sort into contours
		j = 0;
		while (j < nNumEdges)
		{
		    short int		sVtx, sNextVtx;
		    int		lEdge = pContourEL[j++];

		    pnContourSize[nContours] = 1;

		    if (lEdge >= 0)
				sNextVtx = pE[lEdge].endvertex;
		    else
				sNextVtx = pE[-lEdge].startvertex;

		    for (k = j; k < nNumEdges; ++k)
		    {
				lEdge = pContourEL[k];
				if (lEdge >= 0) sVtx = pE[lEdge].startvertex;
				else sVtx = pE[-lEdge].endvertex;
				if (sVtx == sNextVtx)
				{
				    pContourEL[k] = pContourEL[j];
				    pContourEL[j++] = lEdge;
				    ++pnContourSize[nContours];
				    // start looking all over again, account for loop increment
				    k = j - 1;
				    if (lEdge >= 0) sNextVtx = pE[lEdge].endvertex;
				    else sNextVtx = pE[-lEdge].startvertex;
				}
		    }
		    ++nContours;
		}

		if (nContours > 1)
		{
		    printf("Surface %d (%d edges): %d (%d", i, nNumEdges, nContours, pnContourSize[0] );
		    for (j = 1; j < nContours; ++j) printf( ", %d", pnContourSize[j] );
		    printf( ")\n" );
		}

		// for building vertex arrays
		pCurrEpsd->usTessSurf = (int)i;
		pCurrEpsd->pSDL[i].uOffset = pCurrEpsd->uNextSPA;
		pCurrEpsd->pSDL[i].uCount = 0;

		{
		    int		*plNextEdge = pContourEL;
		    double	*pdThisVtx = pdVtx;
		    vec3_t	*pV = pCurrEpsd->pV;

		    // don't bind textures here, causes lists to be host-resident
		    gluTessBeginPolygon( tess, pCurrEpsd );
			for (j = 0; j < nContours; ++j)
			{
			    int		nFirstContourVtx = pdThisVtx - pdVtx;
			    double *pdLastVtx;

			    for (k = 0; k < pnContourSize[j]; ++k)
			    {
					int lEdge = plNextEdge[pnContourSize[j] - 1 - k]; // HACK HACK HACK why is the winding backwards?
					short int sVtx = ((lEdge >= 0) ? pE[lEdge].startvertex : pE[-lEdge].endvertex);

					pdThisVtx[0] = pV[sVtx][0];
					pdThisVtx[1] = pV[sVtx][1];
					pdThisVtx[2] = pV[sVtx][2];
					pdThisVtx += 3;

			    }
			    plNextEdge += pnContourSize[j];
			    gluTessBeginContour( tess );
				pdLastVtx = pdThisVtx - 3;
				pdThisVtx = &pdVtx[nFirstContourVtx];
				for (k = 0; k < pnContourSize[j]; ++k, pdThisVtx += 3)
				{
				    double *pdNextVtx = &pdVtx[nFirstContourVtx + 3 * ((k + 1) % pnContourSize[j])];
				    double dNormTL, dNormNL, dTL[3], dNL[3];
			    
				    dTL[0] = (pdThisVtx[0] - pdLastVtx[0]);
				    dTL[1] = (pdThisVtx[1] - pdLastVtx[1]);
				    dTL[2] = (pdThisVtx[2] - pdLastVtx[2]);
				    dNormTL = 1.0 / sqrt( dTL[0] * dTL[0] + dTL[1] * dTL[1] + dTL[2] * dTL[2] );
				    dTL[0] *= dNormTL;
				    dTL[1] *= dNormTL;
				    dTL[2] *= dNormTL;

				    dNL[0] = (pdNextVtx[0] - pdLastVtx[0]);
				    dNL[1] = (pdNextVtx[1] - pdLastVtx[1]);
				    dNL[2] = (pdNextVtx[2] - pdLastVtx[2]);
				    dNormNL = 1.0 / sqrt( dNL[0] * dNL[0] + dNL[1] * dNL[1] + dNL[2] * dNL[2] );
				    dNL[0] *= dNormNL;
				    dNL[1] *= dNormNL;
				    dNL[2] *= dNormNL;

				    pdLastVtx = pdThisVtx;

				    // skip co-linear points
				    #define COLINEAR_EPSILON 0.00001
				    if ((fabs( dTL[0] - dNL[0] ) <= COLINEAR_EPSILON) &&
					(fabs( dTL[1] - dNL[1] ) <= COLINEAR_EPSILON) && 
					(fabs( dTL[2] - dNL[2] ) <= COLINEAR_EPSILON))
				    {
						++nColinElim;
						continue;
				    }
				    gluTessVertex( tess, pdThisVtx, pdThisVtx );
				}
			    gluTessEndContour( tess );
			}
		    gluTessEndPolygon( tess );
		}
	}

    // setup texgen matricies
    M[ 2] = 0;
    M[ 6] = 0;
    M[10] = 0;
    M[14] = 0;

	M[ 3] = 0;
	M[ 7] = 0;
	M[11] = 0;
	M[15] = 1;
    
    for (i = 0; i < pCurrEpsd->nTG; ++i)
	{
		texgen_t	    *pTG = &pCurrEpsd->pTG[i];
		miptex_t	    *pMipTex = &pCurrEpsd->pMipTex[pTG->texid];
		float			*pfTGM = &pCurrEpsd->pfTGM[i * 8];
		float		    xScale, yScale;

		// setup the display-listed texture matrix
		xScale = 1.0 / pMipTex->width;
		yScale = 1.0 / pMipTex->height;

		pfTGM[0] = M[ 0] = pTG->snrm[0] * xScale;
		pfTGM[1] = M[ 4] = pTG->snrm[1] * xScale;
		pfTGM[2] = M[ 8] = pTG->snrm[2] * xScale;
		pfTGM[3] = M[12] = pTG->soff * xScale;

		pfTGM[4] = M[ 1] = pTG->tnrm[0] * yScale;
		pfTGM[5] = M[ 5] = pTG->tnrm[1] * yScale;
		pfTGM[6] = M[ 9] = pTG->tnrm[2] * yScale;
		pfTGM[7] = M[13] = pTG->toff * yScale;

    }

    bSuccess = TRUE;
Abort:
    if (tess) gluDeleteTess( tess ), tess = NULL;
    if (pdVtx) free( pdVtx );
    if (pnContourSize) free( pnContourSize );
    if (pContourEL) free( pContourEL );
    return( bSuccess );
}

int ClipSegmentToBSP( int nNode, vec3_t *pS, vec3_t *pE, vec3_t *pHit, plane_t **ppPlaneHit, float *pfBias )
{
    int			SP = 0;
    int			nSeg = 0;
    int			i;
    CLIPSTACK		Stack[CLIPSTACK_DEPTH];
    CLIPSTACK		Seg[CLIPSTACK_DEPTH];

    CopyVec3( Stack[SP].S, *pS );
    CopyVec3( Stack[SP].E, *pE );
    Stack[SP].nSplitNode = -99;
    Stack[SP++].nNode = nNode;

    // for now, process everything in the stack
    while (SP > 0)
    {
	vec3_t		S, E;
	int		nSplitNode;

	nNode = Stack[--SP].nNode;
	nSplitNode = Stack[SP].nSplitNode;
	CopyVec3( S, Stack[SP].S );
	CopyVec3( E, Stack[SP].E );

	while (nNode >= 0)
	{
	    bsphullnode_t	*pNode = &pCurrEpsd->pHN[nNode];
	    plane_t		*pPlane = &pCurrEpsd->pP[pNode->planenum];
	    float		fSDist, fEDist;

	    fSDist = DotVec3( pPlane->normal, S ) - pPlane->dist;
	    fEDist = DotVec3( pPlane->normal, E ) - pPlane->dist;

	    //ProgressItem(( "Plane distances start %g end %g", fSDist, fEDist ));

	    // do the start and end points straddle the split plane?
	    if (((fSDist > -CLIP_EPSILON) && (fEDist < CLIP_EPSILON)) ||
		((fSDist < CLIP_EPSILON) && (fEDist > -CLIP_EPSILON)))
	    {
		float fASDist = fabs( fSDist );
		float fAEDist = fabs( fEDist );
		int nSNode, nENode;

		if (fSDist > fEDist)
		{
		    nSNode = pNode->front;
		    nENode = pNode->back;
		}
		else
		{
		    nENode = pNode->front;
		    nSNode = pNode->back;
		}

		// just push end segment so we end up with sorted segs later
		if (SP >= CLIPSTACK_DEPTH)
		{
		    printf( "CLIPSTACK overflow!" );
		    --SP;
		}

		// are we roughly perpendicular to the plane?
		if ((fASDist * fAEDist >= 0) && (fASDist < CLIP_EPSILON) && (fAEDist < CLIP_EPSILON))
		{
		    CopyVec3( Stack[SP].S, S );
		    CopyVec3( Stack[SP].E, E );
		    Stack[SP].nSplitNode = nSplitNode;
		    Stack[SP++].nNode = nENode;
    
		    // continue processing the segment on this side of the plane
		    nNode = nSNode;
		}
		else
		{
		    float fSScale = fAEDist / (fASDist + fAEDist);
		    float fEScale = 1.0 - fSScale;
		    vec3_t M;

		    // find the intersection with the plane
		    M[0] = S[0] * fSScale + E[0] * fEScale;
		    M[1] = S[1] * fSScale + E[1] * fEScale;
		    M[2] = S[2] * fSScale + E[2] * fEScale;
    
		    CopyVec3( Stack[SP].S, M );
		    CopyVec3( Stack[SP].E, E );
		    Stack[SP].nSplitNode = nSplitNode;
		    Stack[SP++].nNode = nENode;

		    // continue processing the start segment
		    nSplitNode = nNode;
		    CopyVec3( E, M );
		    nNode = nSNode;
		}
	    }
	    else // follow the child
	    {
		if ((fSDist + fEDist) > 0) nNode = pNode->front;
		else nNode = pNode->back;
		//ProgressItem(( "New node %d", nNode ));
	    }
	}

	// we can now classify the start segment
	if (nSeg >= CLIPSTACK_DEPTH)
	{
	    printf( "CLIPSEG overflow!" );
	    --nSeg;
	}
	CopyVec3( Seg[nSeg].S, S );
	CopyVec3( Seg[nSeg].E, E );
	Seg[nSeg].nSplitNode = nSplitNode;
	Seg[nSeg++].nNode = nNode;
    }

    nNode = Seg[0].nNode;

    // find first outside->inside transition
    for (i = 0; i < nSeg; ++i)
    {
	if (Seg[i].nNode == CLIP_OUTSIDE) break;
    }

    while (++i < nSeg)
    {
		if (Seg[i].nNode == CLIP_INSIDE)
		{
		    --i;
		    break;
		}
    }

    if (i < nSeg) // we hit something
    {
		bsphullnode_t	*pSplitNode = &pCurrEpsd->pHN[Seg[i].nSplitNode];
		plane_t		*pPlane = &pCurrEpsd->pP[pSplitNode->planenum];

		if (pSplitNode->front < 0)
		{
		    if (pSplitNode->front == CLIP_OUTSIDE) *pfBias = 1;
		    else *pfBias = -1;
		}
		else
		{
		    if (pSplitNode->back == CLIP_OUTSIDE) *pfBias = -1;
		    else *pfBias = 1;
		}


		CopyVec3( *pHit, Seg[i].E );
		*ppPlaneHit = pPlane;
		if (Seg[i].nSplitNode < 0)
		{
//			Notify(( "bad hit plane! seg %d node %d\n", i, Seg[i].nSplitNode ));
			*ppPlaneHit = NULL;
		}
		return( TRUE );
    }
    return( FALSE );
}

int ClipPathToBSP( int nNode, vec3_t *pS, vec3_t *pE )
{
	int			nClips = 3;
	int			bClipped = FALSE;
	plane_t		*pPlane = NULL;
	vec3_t		S;
	vec3_t		Hit;
	float		fBias;

	S[0] = (*pS)[0];
	S[1] = (*pS)[1];
	S[2] = (*pS)[2];

	if ((S[0] == (*pE)[0]) && (S[1] == (*pE)[1]) && (S[2] == (*pE)[2]))
		return( FALSE );

	while (ClipSegmentToBSP( nNode, &S, pE, &Hit, &pPlane, &fBias ))
	{
		bClipped = TRUE;
		if (pPlane && (nClips-- > 0))
		{
			float xe = (*pE)[0], ye = (*pE)[1], ze = (*pE)[2];
			float fEDist = pPlane->normal[0] * xe + pPlane->normal[1]
					* ye + pPlane->normal[2] * ze - pPlane->dist;

			if (fabs( fEDist ) > 100)
			{
				// too noisy, but there's still a problem here...
				//printf( "fEDist %g\n", fEDist );
				fEDist = 10;
			}
			//fEDist -= fBias;
			// project endpoint onto clipping plane and clip to new destination
			(*pE)[0] = xe - fEDist * pPlane->normal[0];
			(*pE)[1] = ye - fEDist * pPlane->normal[1];
			(*pE)[2] = ze - fEDist * pPlane->normal[2];
			 #if defined(DEBUG_CLIPPING)
				ProgressItem(( "(%g,%g,%g) -> (%g,%g,%g), hit (%g,%g,%g)\n",
							xe, ye, ze,
							(*pE)[0], (*pE)[1], (*pE)[2],
							Hit[0], Hit[1], Hit[2] ));
			#endif
			#if 1
			S[0] = Hit[0];
			S[1] = Hit[1];
			S[2] = Hit[2];
			#endif
		}
		else
		{
			(*pE)[0] = Hit[0];
			(*pE)[1] = Hit[1];
			(*pE)[2] = Hit[2];
			break;
		}
		pPlane = NULL;
	}

	return( bClipped );
}

void MoveObject( POBJECT pObj, unsigned int uFrames )
{
	unsigned int	i;
	int				nNode;
	vec3_t			OldHead, OldFeet;

	CopyVec3( OldHead, pObj->Head );
	CopyVec3( OldFeet, pObj->Feet );

	if ((fabs( pObj->fSpeed ) > STOPPED_LIMIT) || (fabs( pObj->fStrafeSpeed ) > STOPPED_LIMIT))
	{
		float fHeadingRad = pObj->fHeading * M_PI / 180.0;
		float dx, dy;
		pObj->Feet[0] += uFrames * (pObj->fSpeed * cos( fHeadingRad )
								+ pObj->fStrafeSpeed
								* cos( fHeadingRad + (90.0 * M_PI / 180.0) ));
		pObj->Feet[1] += uFrames * (pObj->fSpeed * sin( fHeadingRad )
								+ pObj->fStrafeSpeed
								* sin( fHeadingRad + (90.0 * M_PI / 180.0) ));
		dx = pObj->Feet[0] - OldFeet[0];
		dy = pObj->Feet[1] - OldFeet[1];
		pObj->fStep += uFrames * STEP_SPEED_SCALE * sqrt( dx * dx + dy * dy );
		while (pObj->fStep > M_PI)
			pObj->fStep -= (float)M_PI;
	}
	else // when we stop, sink back to standing level
	{
		if (pObj->fStep > M_PI_2)
			pObj->fStep = M_PI - pObj->fStep;
		for (i = 0; i < uFrames; ++i)
			pObj->fStep *= (float)MOTION_DECAY;
	}

	nNode = pCurrEpsd->pH[0].headnode[usClipHull];
	if ((nNode >= 0) && (nNode < pCurrEpsd->nHN))
	{
		int			bFeetOutside;
		vec3_t		Feet;

	// should check for water/slime/lava here...
		for (i = 0; i < uFrames; ++i)
		{
			if (pObj->bGravity)
				pObj->fVertSpeed -= GRAVITY;
			else
				pObj->fVertSpeed *= (float)MOTION_DECAY;
		}
		pObj->Feet[2] += pObj->fVertSpeed;

		bFeetOutside = IsPointOutsideHull( nNode, &pObj->Feet );

		// check for steps
		if (pObj->bOnGround)
		{
			if (!bFeetOutside)
			{
				vec3_t	    OldHip, Hip;
		
				Vec3OpScl( OldHip, OldFeet, *, (1 - HIP_FRACTION) );
				Vec3OpScl( Hip, OldHead, *, HIP_FRACTION );
				Vec3OpVec3( OldHip, OldHip, +, Hip );
				CopyVec3( Hip, pObj->Feet );
				Hip[2] += (float)(pObj->fHeight * HIP_FRACTION);
    
				ClipPathToBSP( nNode, &OldHip, &Hip );
				ClipPathToBSP( nNode, &Hip, &pObj->Feet );
			}
			else
				pObj->bOnGround = FALSE;
		}
		else
		{
			pObj->bOnGround = !bFeetOutside;
			ClipPathToBSP( nNode, &OldFeet, &pObj->Feet );
		}

		CopyVec3( pObj->Head, pObj->Feet );
		CopyVec3( Feet, pObj->Feet );
		pObj->Head[2] += pObj->fHeight;
		Feet[2] += (float)(3 * CLIP_EPSILON);
		if (ClipPathToBSP( nNode, &Feet, &pObj->Head ))
		{
			vec3_t D;
			float d;

			if (pObj->fVertSpeed > 0)
				pObj->fVertSpeed = 0;
			pObj->fSpeed *= (float)MOTION_DECAY;
			pObj->fStrafeSpeed *= (float)MOTION_DECAY;
			Vec3OpVec3( D, pObj->Head, -, pObj->Feet );
			d = DotVec3( D, D );
			pObj->bCrushed = (d < (pObj->fHeight * pObj->fHeight * 0.99));
		}
		else
			pObj->bCrushed = FALSE;

			pObj->bGravity = TRUE;
	}
	else
		pObj->bOnGround = (pObj->Feet[2] < -10000);

	if (pObj->bOnGround)
	{
		// move the head up and down for walking effect
		float fStepCycle = sin( pObj->fStep );
		pObj->Head[2] += STEP_HEIGHT * (fStepCycle * fStepCycle);
		for (i = 0; i < uFrames; ++i)
		{
			pObj->fSpeed *= (float)MOTION_DECAY;
			pObj->fStrafeSpeed *= (float)MOTION_DECAY;
			if (fabs( pObj->fSpeed ) + fabs( pObj->fStrafeSpeed ) > CENTER_SPEED)
				pObj->fPitch *= (float)MOTION_DECAY;
		}
		if (pObj->fVertSpeed < 0)
			pObj->fVertSpeed = 0;
	}

	// setup the view vector
	pObj->View[0] = cos( pObj->fHeading * M_PI / 180.0 );
	pObj->View[1] = sin( pObj->fHeading * M_PI / 180.0 );
	pObj->View[2] = sin( pObj->fPitch * M_PI / 180.0 );
}

int ClipPointToHull( int nNode, vec3_t *p, float radius, plane_t **paPlane )
{
    float	x = (*p)[0], y = (*p)[1], z = (*p)[2];
	int		i = 0, j;
	int	    nCnt = pCurrEpsd->nP;

	for (j = 0; j < nCnt; ++j)
	{
		plane_t *pPlane = &pCurrEpsd->pP[pCurrEpsd->pS[j].planenum];
	
		float fDist;

		fDist = pPlane->normal[0] * x + pPlane->normal[1] * y + pPlane->normal[2] * z - pPlane->dist;

		if (fDist > 0)
		{
			if (fDist < radius)
			{
				paPlane[i++] = pPlane;
			}
			if (i > 255)
				printf("More than 256 planes!\n");

		}
	}
    return( i - 1 );
}

double GetElapsedMilliSecs( void )
{
	static clock_t last = 0;
	clock_t current;
	double dElapsed;

	current = clock();
	dElapsed = (double)(current - last);
	last = current;

	return( dElapsed );
}

int Init( void )
{
// you can play with the fog settings here
#if 0
	float fFogColor[] = { 0.5f, 0.5f, 0.5f, 1.0f };

	glEnable( GL_FOG );
	glFogi( GL_FOG_MODE, GL_EXP );
	glFogf( GL_FOG_START, 0.1f );
	glFogfv( GL_FOG_COLOR, fFogColor );
	glFogf( GL_FOG_DENSITY, 0.005f );

#endif
	bDrawTexture = TRUE;

	pGLVendor = (char *)glGetString(GL_VENDOR);
	pGLRenderer = (char *)glGetString(GL_RENDERER);
    glEnable( GL_CULL_FACE );
    glEnable( GL_DEPTH_TEST );
	glDrawBuffer( GL_BACK );
// disabling these may speed up some systems, does not effect 3DFx Voodoo
//	glDisable( GL_TEXTURE_COORD_ARRAY_EXT );
//	glDisable( GL_NORMAL_ARRAY_EXT );
//	glDisable( GL_VERTEX_ARRAY_EXT );
//	glDisable( GL_DITHER );
//	glDisable( GL_LIGHTING );
//	glDisable( GL_ALPHA_TEST );

	glEnable( GL_TEXTURE_2D );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );

    if (!BuildTextures())
		BAIL(( "Failed BuildTextures!" ));
    if (!TesselateEpisode())
		BAIL(( "Failed TesselateEpisode!" ));
    if (!CalcTexCoords())
		goto Abort;

	// set the initial Player settings
	// these should come from the map file
	Player.fHeading = 90;
	Player.fPitch = 0;
	Player.fRoll = 0;
	Player.fRadius = 20;
	Player.fHeight = 24;
	Player.Feet[0] = 540;	// x position
	Player.Feet[1] = 260;	// y position
	Player.Feet[2] = 120;	// z position( up )
	Player.Head[0] = Player.Feet[0];
	Player.Head[1] = Player.Feet[1];
	Player.Head[2] = Player.Feet[2] + Player.fHeight;
	Player.fSpeed = 0;
	Player.fStrafeSpeed = 0;
	Player.fVertSpeed = 0;
	Player.fStep = 0;
	Player.bGravity = TRUE;
	Player.bOnGround = TRUE;
	Player.bInLiquid = FALSE;

	return(TRUE);
Abort:
	return(FALSE);
}

#ifdef WIN32
void winkeyboard( void )
{
	int			usOldHull = usHull;
	static int	toggleT = FALSE, toggleH = FALSE;
	float		fAcceleration, fTurnSpeed;

	if(GetAsyncKeyState( VK_ESCAPE ))
		exit(0);

	if(GetAsyncKeyState( 'T' ))
	{
		if (!toggleT)
		{
			bDrawTexture ^= TRUE;
			toggleT = TRUE;
		}
	}
	else
		toggleT = FALSE;


	if(GetAsyncKeyState( 'H' ))
	{
		if (!toggleH)
		{
			bDrawHulls ^= TRUE;
			toggleH = TRUE;
		}
	}
	else
		toggleH = FALSE;

	if(GetAsyncKeyState( VK_SHIFT ))
	{
	    fAcceleration = RUNNING_ACCEL;
	    fTurnSpeed = RUNNING_TURN_SPEED;
	}
	else
	{
		fAcceleration = WALKING_ACCEL;
		fTurnSpeed = WALKING_TURN_SPEED;
	}

	if(GetAsyncKeyState( VK_LEFT ))
	{
		Player.fHeading += fTurnSpeed * uElapsedFrames / 2;
		if (Player.fHeading > 180.0)
			Player.fHeading -= 360.0;
	}

	if(GetAsyncKeyState( VK_RIGHT ))
	{
		Player.fHeading -= fTurnSpeed * uElapsedFrames / 2;
		if (Player.fHeading < 180.0)
			Player.fHeading += 360.0;
	}

	if(GetAsyncKeyState( VK_UP ))
		Player.fSpeed += fAcceleration;
    
	if(GetAsyncKeyState( VK_DOWN ))
		Player.fSpeed -= fAcceleration;

	if(GetAsyncKeyState( VK_END ))
		Player.fVertSpeed = JUMP_ACCEL;

	if(GetAsyncKeyState( VK_PRIOR ))//Page Up
	{
		if (Player.fPitch < 89.0)
			Player.fPitch += fTurnSpeed * uElapsedFrames / 2;
	}

	if(GetAsyncKeyState( VK_NEXT ))//Page Down
	{
		if (Player.fPitch > -89.0)
			Player.fPitch -= fTurnSpeed * uElapsedFrames / 2;
	}
}
#else
static void special(int key, int x, int y)
{
	float		fAcceleration, fTurnSpeed;
    
	fAcceleration = WALKING_ACCEL;
	fTurnSpeed = WALKING_TURN_SPEED;
//	    fAcceleration = RUNNING_ACCEL;
//	    fTurnSpeed = RUNNING_TURN_SPEED;

	switch (key) {
		case GLUT_KEY_LEFT:
			Player.fHeading += fTurnSpeed;
			if (Player.fHeading > 180.0)
				Player.fHeading -= 360.0;
			break;
		case GLUT_KEY_RIGHT:
			Player.fHeading -= fTurnSpeed;
			if (Player.fHeading < 180.0)
				Player.fHeading += 360.0;
			break;
		case GLUT_KEY_UP:
			Player.fSpeed += fAcceleration;
			break;
		case GLUT_KEY_DOWN:
			Player.fSpeed -= fAcceleration;
			break;
		case GLUT_KEY_END:
			Player.fVertSpeed = JUMP_ACCEL;
			break;
		case GLUT_KEY_PAGE_UP:
			if (Player.fPitch < 89.0)
				Player.fPitch += fTurnSpeed;
			break;
		case GLUT_KEY_PAGE_DOWN:
			if (Player.fPitch > -89.0)
				Player.fPitch -= fTurnSpeed;
			break;
	}
}

static void keyboard(unsigned char key, int x, int y)
{
	switch (key) {
		case 27:
			exit(0);
			break;
		case 't':
			bDrawTexture ^= TRUE;
			break;
		case 'h':
			bDrawHulls ^= TRUE;
			break;
	}
}

#endif // WIN32

void UpdateLevel( void )
{
	static int	nFrames = 0;
	static int	nUpdate = 0;

	dRealTime += GetElapsedMilliSecs();
	uElapsedFrames = (unsigned int)((dRealTime - dSimTime) / SIM_TICK);

	// adjust fps update to about 2 times per second
	nUpdate = max( 2, (int)((fps/2 + nUpdate)/2));

	if (++nFrames >= nUpdate)
    {
		static double dLastTime = 0;
		fps = 1000 / (dRealTime - dLastTime) * nUpdate;
		printf("fps: %6.2f \n", fps );
		dLastTime = dRealTime;
		nFrames = 0;
    }

	dSimTime += uElapsedFrames * SIM_TICK;
	uFrame += uElapsedFrames;

#ifdef WIN32
	winkeyboard();
#endif
	MoveObject( &Player, uElapsedFrames );
	RenderLevel();
}

int main( int argc, char *argv[] )
{

    Notify((EGLE_VERSION));
	Notify(("\n"));
	Notify(("by Phil Frisbie, Jr.\n"));

	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ALPHA);
	glutInitWindowSize( W, H);
	glutCreateWindow(EGLE_VERSION);
	if (argc > 1)
		pszMap = argv[1];

	if (!LoadFiles())
		goto Abort;

	if (!Init())
		goto Abort;
	SetViews();

	glutDisplayFunc(UpdateLevel);
	glutIdleFunc(UpdateLevel);
#ifndef WIN32
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(special);
#endif
	// init the function
    GetElapsedMilliSecs();
    dRealTime = dSimTime = 0;

	glutMainLoop();

Abort:
    return( 0 );
}

