#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include "egle.h"


unsigned int Palette[256];

int load_file (char *filename, void **bufferptr)
{
	int f;
	int length;
	void *buffer;
	struct stat	st;

	f = open(filename, O_RDONLY);
	if(f == -1)
	{
		printf("Cannot open file %s\n", filename);
		exit(2);
	}
	stat( filename, &st );
	length = st.st_size;
	lseek(f, 0, SEEK_SET);
	buffer = malloc (length);
	read(f, buffer, length);
	close(f);

	*bufferptr = buffer;
	return length;
}

short LittleShort(short l)
{
	return l;
}

int LittleLong(int l)
{
	return l;
}

float LittleFloat(float l)
{
	return l;
}

void load_palette(void)
{
	int				f = 0;
	int				i;
	unsigned char	*temp_pal;
    int			    nGamma[256];
    double		    dGamma = 0.6;

	load_file("palette.lmp", (void **)&temp_pal);
	for(i=0;i<256;i++)
	{
#ifdef ABGR
		Palette[i] = ((((LONG)temp_pal[0])) | (((LONG)temp_pal[1]) << 8) | (((LONG)temp_pal[2]) << 16) | (((i != 255) ? 0xff : 0) << 24));
#else
		Palette[i] = ((((LONG)temp_pal[0]))<<24 | (((LONG)temp_pal[1]) << 16) | (((LONG)temp_pal[2]) << 8) | (((i != 255) ? 0xff : 0)));
#endif
		temp_pal += 3;
	}
    for (i = 0 ; i < 256; ++i)
    {
		int inf = 255 * pow( (i + 0.5) / 255.5, dGamma ) + 0.5;
		if (inf < 0) inf = 0;
		else if (inf > 255) inf = 255;
		nGamma[i] = inf;
    }

    for (i = 0 ; i < 256; ++i)
    {
		unsigned int ulRGBA = Palette[i];
#ifdef ABGR
		unsigned int ulR = nGamma[(ulRGBA) & 0xff];
		unsigned int ulG = nGamma[(ulRGBA >> 8) & 0xff];
		unsigned int ulB = nGamma[(ulRGBA >> 16) & 0xff];
		Palette[i] = (ulR) | (ulG << 8) | (ulB << 16) | ((ulRGBA & 0xff) <<24);
#else
		unsigned int ulR = nGamma[(ulRGBA >> 24) & 0xff];
		unsigned int ulG = nGamma[(ulRGBA >> 16) & 0xff];
		unsigned int ulB = nGamma[(ulRGBA >>  8) & 0xff];
		Palette[i] = (ulR << 24) | (ulG << 16) | (ulB << 8) | (ulRGBA & 0xff);
#endif
    }

	temp_pal -= 768;
	free(temp_pal);
}

int LoadMap( char * pszName )
{
    char			*buffer;
    dheader_t		*pSrcHead;
    dheader_t		*pDstHead;
    vec3_t			*pSrcV;
    edge_t			*pSrcE;
    int				*pSrcEL;
    plane_t			*pSrcP;
    surface_t		*pSrcS;
    u_short			*pSrcLS;
    texgen_t		*pSrcTG;
    bsphull_t		*pSrcH;
    bsphullnode_t	*pSrcHN;
    bspnode_t		*pSrcN;
    bspleaf_t		*pSrcL;
    mipheader_t		*pSrcMipHead;
    miptex_t		*pSrcMipTex;
    int				nEntries, i, j;
    int				nMinSurf, nMaxSurf, nMinSurfNum, nMaxSurfNum;
    int				bSuccess = FALSE;

    Progress(( "\nLoading %s\n", pszName ));

	load_file( pszName, (void **)&buffer);

	if (pCurrEpsd) DestroyEpisode( pCurrEpsd ), pCurrEpsd = NULL;
    pCurrEpsd = CreateEpisode();
    if (!pCurrEpsd) BAIL(( "Failed to create episode!" ));

    pSrcHead = (dheader_t *)buffer;
    pDstHead = &pCurrEpsd->Head;
    pDstHead->version = LittleLong( pSrcHead->version );

    if (pDstHead->version != BSP_VERSION)
		Notify(( "Quake BSP file version mismatch (want %d got %d)\n", BSP_VERSION, pDstHead->version ));
	pDstHead->entities.offset = LittleLong(pSrcHead->entities.offset);
	pDstHead->entities.size = LittleLong(pSrcHead->entities.size);
	pDstHead->planes.offset = LittleLong(pSrcHead->planes.offset);
	pDstHead->planes.size = LittleLong(pSrcHead->planes.size);
	pDstHead->miptex.offset = LittleLong(pSrcHead->miptex.offset);
	pDstHead->miptex.size = LittleLong(pSrcHead->miptex.size);
	pDstHead->vertices.offset = LittleLong(pSrcHead->vertices.offset);
	pDstHead->vertices.size = LittleLong(pSrcHead->vertices.size);
	pDstHead->visilist.offset = LittleLong(pSrcHead->visilist.offset);
	pDstHead->visilist.size = LittleLong(pSrcHead->visilist.size);
	pDstHead->nodes.offset = LittleLong(pSrcHead->nodes.offset);
	pDstHead->nodes.size = LittleLong(pSrcHead->nodes.size);
	pDstHead->texgen.offset = LittleLong(pSrcHead->texgen.offset);
	pDstHead->texgen.size = LittleLong(pSrcHead->texgen.size);
	pDstHead->surfaces.offset = LittleLong(pSrcHead->surfaces.offset);
	pDstHead->surfaces.size = LittleLong(pSrcHead->surfaces.size);
	pDstHead->lightmaps.offset = LittleLong(pSrcHead->lightmaps.offset);
	pDstHead->lightmaps.size = LittleLong(pSrcHead->lightmaps.size);
	pDstHead->boundnodes.offset = LittleLong(pSrcHead->boundnodes.offset);
	pDstHead->boundnodes.size = LittleLong(pSrcHead->boundnodes.size);
	pDstHead->leaves.offset = LittleLong(pSrcHead->leaves.offset);
	pDstHead->leaves.size = LittleLong(pSrcHead->leaves.size);
	pDstHead->lstsurf.offset = LittleLong(pSrcHead->lstsurf.offset);
	pDstHead->lstsurf.size = LittleLong(pSrcHead->lstsurf.size);
	pDstHead->edges.offset = LittleLong(pSrcHead->edges.offset);
	pDstHead->edges.size = LittleLong(pSrcHead->edges.size);
	pDstHead->lstedges.offset = LittleLong(pSrcHead->lstedges.offset);
	pDstHead->lstedges.size = LittleLong(pSrcHead->lstedges.size);
	pDstHead->hulls.offset = LittleLong(pSrcHead->hulls.offset);
	pDstHead->hulls.size = LittleLong(pSrcHead->hulls.size);

    //
    // entities
    //

    Progress(( "entities at 0x%08x (%d bytes)\n", pDstHead->entities.offset, pDstHead->entities.size ));

    //
    // verticies
    //
    
	nEntries = pDstHead->vertices.size / sizeof(pSrcV[0]);
	Progress(( "verticies at 0x%08x (%d entries)\n", pDstHead->vertices.offset, nEntries ));
	pCurrEpsd->pV = malloc( sizeof(pCurrEpsd->pV[0]) * nEntries );
	if (!pCurrEpsd->pV) BAIL(( "Out of memory!" ));

	pSrcV = (vec3_t *)(&buffer[pDstHead->vertices.offset]);
	for (i = 0; i < nEntries; ++i)
	{
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pV[i][j] = LittleFloat(pSrcV[i][j]);
	}
	pCurrEpsd->nV = nEntries;

	//
	// edges
	//

	nEntries = pDstHead->edges.size / sizeof(pSrcE[0]);
	Progress(( "edges at 0x%08x (%d entries)\n", pDstHead->edges.offset, nEntries ));
	pCurrEpsd->pE = malloc( sizeof(pCurrEpsd->pE[0]) * nEntries );
	if (!pCurrEpsd->pE) BAIL(( "Out of memory!" ));

	pSrcE = (edge_t *)(&buffer[pDstHead->edges.offset]);
	for (i = 0; i < nEntries; ++i)
	{
		pCurrEpsd->pE[i].startvertex = LittleShort(pSrcE[i].startvertex);
		pCurrEpsd->pE[i].endvertex = LittleShort(pSrcE[i].endvertex);
    }
	pCurrEpsd->nE = nEntries;

	//
	// edgelists
	//
    
	nEntries = pDstHead->lstedges.size / sizeof(pSrcEL[0]);
	Progress(( "lstedges at 0x%08x (%d entries)\n", pDstHead->lstedges.offset, nEntries ));
	pCurrEpsd->pEL = malloc( sizeof(pCurrEpsd->pEL[0]) * nEntries );
	if (!pCurrEpsd->pEL) BAIL(( "Out of memory!" ));

	pSrcEL = (int *)(&buffer[pDstHead->lstedges.offset]);
	for (i = 0; i < nEntries; ++i)
		pCurrEpsd->pEL[i] = LittleLong(pSrcEL[i]);
	pCurrEpsd->nEL = nEntries;

	//
	// planes
	//

	nEntries = pDstHead->planes.size / sizeof(pSrcP[0]);
	Progress(( "planes at 0x%08x (%d entries)\n", pDstHead->planes.offset, nEntries ));
	pCurrEpsd->pP = malloc( sizeof(pCurrEpsd->pP[0]) * nEntries );
	if (!pCurrEpsd->pP) BAIL(( "Out of memory!" ));

	pSrcP = (plane_t *)(&buffer[pDstHead->planes.offset]);
	for (i = 0; i < nEntries; ++i)
	{
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pP[i].normal[j] = LittleFloat(pSrcP[i].normal[j]);
		pCurrEpsd->pP[i].dist = LittleFloat(pSrcP[i].dist);
		pCurrEpsd->pP[i].type = LittleLong(pSrcP[i].type);
	}
	pCurrEpsd->nP = nEntries;

	//
	// surfaces
	//

	nEntries = pDstHead->surfaces.size / sizeof(pSrcS[0]);
	Progress(( "surfaces at 0x%08x (%d entries)\n", pDstHead->surfaces.offset, nEntries ));
	pCurrEpsd->pS = malloc( sizeof(pCurrEpsd->pS[0]) * nEntries );

	pSrcS = (surface_t *)(&buffer[pDstHead->surfaces.offset]);
	pCurrEpsd->nMaxEdgesPerSurface = 0;
	for (i = 0; i < nEntries; ++i)
	{
		pCurrEpsd->pS[i].planenum = LittleShort(pSrcS[i].planenum);
		pCurrEpsd->pS[i].side = LittleShort(pSrcS[i].side);
		pCurrEpsd->pS[i].firstedge = LittleLong(pSrcS[i].firstedge);
		pCurrEpsd->pS[i].numedge = LittleShort(pSrcS[i].numedge);
		pCurrEpsd->pS[i].texgenid = LittleShort(pSrcS[i].texgenid);
		pCurrEpsd->pS[i].lightmap = LittleLong(pSrcS[i].lightmap);
		pCurrEpsd->pS[i].lightstyle[0] = pSrcS[i].lightstyle[0];
		pCurrEpsd->pS[i].lightstyle[1] = pSrcS[i].lightstyle[1];
		pCurrEpsd->pS[i].lightstyle[2] = pSrcS[i].lightstyle[2];
		pCurrEpsd->pS[i].lightstyle[3] = pSrcS[i].lightstyle[3];

		if (pCurrEpsd->nMaxEdgesPerSurface < pCurrEpsd->pS[i].numedge)
			pCurrEpsd->nMaxEdgesPerSurface = pCurrEpsd->pS[i].numedge;
	}
	pCurrEpsd->nS = nEntries;

	//
	// surface lists
	//

	nEntries = pDstHead->lstsurf.size / sizeof(pSrcLS[0]);
	Progress(( "lstsurf at 0x%08x (%d entries)\n", pDstHead->lstsurf.offset, nEntries ));
	pCurrEpsd->pLS = malloc( sizeof(pCurrEpsd->pLS[0]) * nEntries );
	// this array is a scratch area used during rendering
	pCurrEpsd->pTmpLS = malloc( sizeof(pCurrEpsd->pTmpLS[0]) * nEntries );
	if (!pCurrEpsd->pLS || !pCurrEpsd->pTmpLS) BAIL(( "Out of memory!" ));

	pSrcLS = (u_short *)(&buffer[pDstHead->lstsurf.offset]);
	for (i = 0; i < nEntries; ++i)
		pCurrEpsd->pLS[i] = LittleShort(pSrcLS[i]);
	pCurrEpsd->nLS = nEntries;

	//
	// visibility lists
	//

	nEntries = pDstHead->visilist.size / sizeof(pCurrEpsd->pVL[0]);
	Progress(( "visilist at 0x%08x (%d entries)\n", pDstHead->visilist.offset, nEntries ));
	pCurrEpsd->pVL = (u_char *)(&buffer[pDstHead->visilist.offset]);
	pCurrEpsd->nVL = nEntries;

	//
	// texgen
	//

	nEntries = pDstHead->texgen.size / sizeof(pSrcTG[0]);
	Progress(( "texgens at 0x%08x (%d entries)\n", pDstHead->texgen.offset, nEntries ));
	pCurrEpsd->pTG = malloc( sizeof(pCurrEpsd->pTG[0]) * nEntries );
	if (!pCurrEpsd->pTG) BAIL(( "Out of memory!" ));

	pSrcTG = (texgen_t *)(&buffer[pDstHead->texgen.offset]);
	for (i = 0; i < nEntries; ++i)
	{
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pTG[i].snrm[j] = LittleFloat(pSrcTG[i].snrm[j]);
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pTG[i].tnrm[j] = LittleFloat(pSrcTG[i].tnrm[j]);
		pCurrEpsd->pTG[i].soff = LittleFloat(pSrcTG[i].soff);
		pCurrEpsd->pTG[i].toff = LittleFloat(pSrcTG[i].toff);
		pCurrEpsd->pTG[i].texid = LittleLong(pSrcTG[i].texid);
		pCurrEpsd->pTG[i].flag = LittleLong(pSrcTG[i].flag);

	}
	pCurrEpsd->nTG = nEntries;

	//
	// bsp hulls
	//

	nEntries = pDstHead->hulls.size / sizeof(pSrcH[0]);
	Progress(( "hulls at 0x%08x (%d bytes, %d entries)\n", pDstHead->hulls.offset, pDstHead->hulls.size, nEntries ));
	pCurrEpsd->pH = malloc( sizeof(pCurrEpsd->pH[0]) * nEntries );
	if (!pCurrEpsd->pH) BAIL(( "Out of memory!" ));

	pSrcH = (bsphull_t *)(&buffer[pDstHead->hulls.offset]);
	for (i = 0; i < nEntries; ++i)
	{
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pH[i].box.min[j] = LittleFloat(pSrcH[i].box.min[j]);
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pH[i].box.max[j] = LittleFloat(pSrcH[i].box.max[j]);
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pH[i].origin[j] = LittleFloat(pSrcH[i].origin[j]);
		pCurrEpsd->pH[i].headnode[0] = LittleLong(pSrcH[i].headnode[0]);
		pCurrEpsd->pH[i].headnode[1] = LittleLong(pSrcH[i].headnode[1]);
		pCurrEpsd->pH[i].headnode[2] = LittleLong(pSrcH[i].headnode[2]);
		pCurrEpsd->pH[i].headnode[3] = LittleLong(pSrcH[i].headnode[3]);
		pCurrEpsd->pH[i].visleaves = LittleLong(pSrcH[i].visleaves);
		pCurrEpsd->pH[i].firstsurf = LittleLong(pSrcH[i].firstsurf);
		pCurrEpsd->pH[i].numsurf = LittleLong(pSrcH[i].numsurf);

	}
	pCurrEpsd->nH = nEntries;

	//
	// bsp hull nodes
	//

	nEntries = pDstHead->boundnodes.size / sizeof(pSrcHN[0]);
	Progress(( "boundnodes at 0x%08x (%d bytes, %d entries)\n", pDstHead->boundnodes.offset, pDstHead->boundnodes.size, nEntries ));
	pCurrEpsd->pHN = malloc( sizeof(pCurrEpsd->pHN[0]) * nEntries );
	if (!pCurrEpsd->pHN) BAIL(( "Out of memory!" ));

	pSrcHN = (bsphullnode_t *)(&buffer[pDstHead->boundnodes.offset]);
	for (i = 0; i < nEntries; ++i)
	{
		pCurrEpsd->pHN[i].planenum = LittleLong(pSrcHN[i].planenum);
		pCurrEpsd->pHN[i].front = LittleShort(pSrcHN[i].front);
		pCurrEpsd->pHN[i].back = LittleShort(pSrcHN[i].back);

	}
	pCurrEpsd->nHN = nEntries;

	//
	// bsp render nodes
	//

	nEntries = pDstHead->nodes.size / sizeof(pSrcN[0]);
	Progress(( "nodes at 0x%08x (%d bytes, %d entries)\n", pDstHead->nodes.offset, pDstHead->nodes.size, nEntries ));
	pCurrEpsd->pN = malloc( sizeof(pCurrEpsd->pN[0]) * nEntries );
	if (!pCurrEpsd->pN) BAIL(( "Out of memory!" ));

	pSrcN = (bspnode_t *)(&buffer[pDstHead->nodes.offset]);
	for (i = 0; i < nEntries; ++i)
	{
		pCurrEpsd->pN[i].planenum = LittleLong(pSrcN[i].planenum);
		pCurrEpsd->pN[i].front = LittleShort(pSrcN[i].front);
		pCurrEpsd->pN[i].back = LittleShort(pSrcN[i].back);
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pN[i].sbox.min[j] = LittleFloat(pSrcN[i].sbox.min[j]);
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pN[i].sbox.max[j] = LittleFloat(pSrcN[i].sbox.max[j]);
		pCurrEpsd->pN[i].firstsurf = LittleShort(pSrcN[i].firstsurf);
		pCurrEpsd->pN[i].numsurf = LittleShort(pSrcN[i].numsurf);

	}
	pCurrEpsd->nN = nEntries;

	//
	// bsp leaves
	//

	nEntries = pDstHead->leaves.size / sizeof(pSrcL[0]);
	Progress(( "leaves at 0x%08x (%d bytes, %d entries)\n", pDstHead->leaves.offset, pDstHead->leaves.size, nEntries ));
	pCurrEpsd->pL = malloc( sizeof(pCurrEpsd->pL[0]) * nEntries );
	if (!pCurrEpsd->pL) BAIL(( "Out of memory!" ));

	pSrcL = (bspleaf_t *)(&buffer[pDstHead->leaves.offset]);
	for (i = 0; i < nEntries; ++i)
	{
		pCurrEpsd->pL[i].type = LittleLong(pSrcL[i].type);
		pCurrEpsd->pL[i].vislist = LittleLong(pSrcL[i].vislist);
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pL[i].sbox.min[j] = LittleFloat(pSrcL[i].sbox.min[j]);
		for (j=0 ; j<3 ; j++)
			pCurrEpsd->pL[i].sbox.max[j] = LittleFloat(pSrcL[i].sbox.max[j]);
		pCurrEpsd->pL[i].firstsurf = LittleShort(pSrcL[i].firstsurf);
		pCurrEpsd->pL[i].numsurf = LittleShort(pSrcL[i].numsurf);
		pCurrEpsd->pL[i].ambientsnd[0] = pSrcL[i].ambientsnd[0];
		pCurrEpsd->pL[i].ambientsnd[1] = pSrcL[i].ambientsnd[1];
		pCurrEpsd->pL[i].ambientsnd[2] = pSrcL[i].ambientsnd[2];
		pCurrEpsd->pL[i].ambientsnd[3] = pSrcL[i].ambientsnd[3];

		if (i == 0)
		{
			nMinSurf = nMaxSurf = pCurrEpsd->pL[i].firstsurf;
			nMinSurfNum = nMaxSurfNum = pCurrEpsd->pL[i].numsurf;
		}
		else
		{
			if (nMinSurf > pCurrEpsd->pL[i].firstsurf) nMinSurf = pCurrEpsd->pL[i].firstsurf;
			if (nMaxSurf < pCurrEpsd->pL[i].firstsurf) nMaxSurf = pCurrEpsd->pL[i].firstsurf;
			if (nMinSurfNum > pCurrEpsd->pL[i].numsurf) nMinSurfNum = pCurrEpsd->pL[i].numsurf;
			if (nMaxSurfNum < pCurrEpsd->pL[i].numsurf) nMaxSurfNum = pCurrEpsd->pL[i].numsurf;
		}

	}
	pCurrEpsd->nL = nEntries;

	//
	// miptex
	//

	pSrcMipHead = (mipheader_t *)&(buffer[pDstHead->miptex.offset]);
	nEntries = LittleLong(pSrcMipHead->numtex);
	Progress(( "miptex at 0x%08x (%d bytes, %d entries)\n", pDstHead->miptex.offset, pDstHead->miptex.size, nEntries ));
	pCurrEpsd->pMipHead = malloc( sizeof(pCurrEpsd->pMipHead[0]) + sizeof(pCurrEpsd->pMipHead->offset[0]) * nEntries );
	pCurrEpsd->pMipTex = malloc( sizeof(pCurrEpsd->pMipTex[0]) * nEntries );
	pCurrEpsd->ppucTexture = malloc( sizeof(pCurrEpsd->ppucTexture[0]) * nEntries );
	if (!pCurrEpsd->pMipHead || !pCurrEpsd->pMipTex || !pCurrEpsd->ppucTexture) BAIL(( "Out of memory!" ));
		pCurrEpsd->pMipHead->numtex = nEntries;

	for (i = 0; i < nEntries; ++i)
	{
		int lOffset;

		pCurrEpsd->pMipHead->offset[i] = LittleLong(pSrcMipHead->offset[i]);

		lOffset = pDstHead->miptex.offset + pCurrEpsd->pMipHead->offset[i];
		pSrcMipTex = (miptex_t *)&(buffer[lOffset]);
		strncpy( pCurrEpsd->pMipTex[i].name, pSrcMipTex->name, sizeof(pCurrEpsd->pMipTex[i].name) );
		pCurrEpsd->pMipTex[i].width = LittleLong(pSrcMipTex->width);
		pCurrEpsd->pMipTex[i].height = LittleLong(pSrcMipTex->height);
		pCurrEpsd->pMipTex[i].offset1 = LittleLong(pSrcMipTex->offset1);
		pCurrEpsd->pMipTex[i].offset2 = LittleLong(pSrcMipTex->offset2);
		pCurrEpsd->pMipTex[i].offset4 = LittleLong(pSrcMipTex->offset4);
		pCurrEpsd->pMipTex[i].offset8 = LittleLong(pSrcMipTex->offset8);

		lOffset += pCurrEpsd->pMipTex[i].offset1;
		pCurrEpsd->ppucTexture[i] = (unsigned char *)&(buffer[lOffset]);
	}

	//
	// sort the visible surfaces in hull 0 by texture id
	//

	// surfaces sorted by texture id
	pCurrEpsd->pSortedS = malloc( sizeof(pCurrEpsd->pSortedS[0]) * pCurrEpsd->nS );
	if (!pCurrEpsd->pSortedS ) BAIL(( "Out of memory!" ));

	for (i = 0; i < pCurrEpsd->nS; ++i)
		pCurrEpsd->pSortedS[i] = i;

	for (i = 0; i < pCurrEpsd->nH; ++i)
	{
		qsort( &pCurrEpsd->pSortedS[pCurrEpsd->pH[i].firstsurf],
			pCurrEpsd->pH[i].numsurf, sizeof(pCurrEpsd->pSortedS[0]), CmpSurfTex );
	}

	bSuccess = TRUE;

Abort:
	if (!bSuccess)
	{
		if (pCurrEpsd) DestroyEpisode( pCurrEpsd ), pCurrEpsd = NULL;
	}
	return( bSuccess );
}

int LoadFiles( void )
{
	load_palette();
    if (!LoadMap( pszMap )) return(FALSE);
    return(TRUE);
}
