#include "egle.h"

void DrawSurface( int i )
{
	unsigned int uCount = pCurrEpsd->pSDL[i].uCount;
	PPRIMDESC pPrim = &pCurrEpsd->pSPA[pCurrEpsd->pSDL[i].uOffset];
	do {
		short int sCount = pPrim->p.sCount;
		glBegin( (pPrim++)->p.sMode );
		do {
			glTexCoord2fv( pPrim->c.t );
			glVertex3fv( (pPrim++)->c.v );
			} while (--sCount) ;
			glEnd();
	} while (--uCount) ;
}  

void BeginOverlay( void )
{
	glMatrixMode( GL_PROJECTION );
	glPushMatrix();

	glLoadIdentity();
	gluOrtho2D( 0.0, W, 0.0, H );
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();

	glLoadIdentity();
}

void EndOverlay( void )
{
	glPopMatrix();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
}

void DrawMessage( int x, int y, char *pszFormat, ... )
{
	char	    szBuff[1024];
	va_list	    vArgs;
	int			len,i;

	va_start( vArgs, pszFormat );
	vsprintf( szBuff, pszFormat, vArgs );
	va_end( vArgs );
	len=(int)strlen(szBuff);

	glColor3f(1.0f, 1.0f, 1.0f);
	glRasterPos2f( x, y );

	for(i=0;i<len;i++)
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13,szBuff[i]);

}

int CmpSurfTex( const void *p1, const void *p2 )
{
	unsigned short int	s1 = *(unsigned short int *)p1;
	unsigned short int	s2 = *(unsigned short int *)p2;
	surface_t			*pS1 = &pCurrEpsd->pS[s1];
	surface_t			*pS2 = &pCurrEpsd->pS[s1];
	texgen_t			*pTG1 = &pCurrEpsd->pTG[pS1->texgenid];
	texgen_t			*pTG2 = &pCurrEpsd->pTG[pS2->texgenid];
	int					nDiff;
    
	nDiff = (int)(pTG2->texid - pTG1->texid);
	if (!nDiff) nDiff = s2 - s1;
	return( nDiff );
}

int QuickRejectLeaf( bspleaf_t *pLeaf, PVIEW pView )
{
	vec3_t	    v;
	float	    d;

	// pick bbox corner most likely to be visible along view direction
	v[0] = (pView->View[0] > 0) ? pLeaf->sbox.max[0] : pLeaf->sbox.min[0];
	v[1] = (pView->View[1] > 0) ? pLeaf->sbox.max[1] : pLeaf->sbox.min[1];
	v[2] = (pView->View[2] > 0) ? pLeaf->sbox.max[2] : pLeaf->sbox.min[2];
	d = (v[0] - pView->pObject->Head[0]) * pView->View[0]
		+ (v[1] - pView->pObject->Head[1]) * pView->View[1]
		+ (v[2] - pView->pObject->Head[2]) * pView->View[2];

	return( d < 0 );
}

void DrawSurfaceList( int * pVS, int nVS, PVIEW pView )
{
	int			usLastSurf;
	int		    i;
	surface_t	*pS = pCurrEpsd->pS;
	int		    nS = pCurrEpsd->nS;
	int		    lLastBoundTex = -1;
	int		    lSkyTex = -1;
	texgen_t	*pOldTG = NULL;

    glColor4ub( 0xff, 0xff, 0xff, 0xff );
  
    if (bDrawTexture)
    {
		int	    bSky = FALSE;

		glEnable( GL_TEXTURE_2D );
		glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
		glShadeModel( GL_FLAT );
		glMatrixMode( GL_TEXTURE );

		for (i = 0, usLastSurf = 0xffff; i < nVS; ++i)
		{
		    int			usSurf = pVS[i];
		    texgen_t	*pTG;

		    if (usSurf == usLastSurf) continue;
		    usLastSurf = usSurf;
		    pTG = &pCurrEpsd->pTG[pS[usSurf].texgenid];
		    if (pOldTG != pTG)
		    {
				pOldTG = pTG;
		    }

		    if (lLastBoundTex != pTG->texid)
		    {
				unsigned int uTex = 0;
				lLastBoundTex = pTG->texid;
				bSky = FALSE;

				if (pTG->flag)
				{
				    char *pszName = pCurrEpsd->pMipTex[lLastBoundTex].name;
					if (pszName[0] == '*')
				    {
						uTex = AnimTexIndex( uFrame );
				    }
				    else if ((pszName[0] == 's') && (pszName[1] == 'k') && (pszName[2] == 'y'))
				    {

						bSky = TRUE;
						lSkyTex = lLastBoundTex;
						continue;
				    }
				}

				glBindTexture( GL_TEXTURE_2D, pCurrEpsd->puTexObj[pCurrEpsd->puTexObjBase[lLastBoundTex] + uTex] );
		    }

		    if (!bSky) DrawSurface( usSurf );
		}

		// draw sky textures
		if (lSkyTex != -1)
		{
		    float fSpeedScale;

		    // do background sky texture
		    glBindTexture( GL_TEXTURE_2D, pCurrEpsd->puTexObj[pCurrEpsd->puTexObjBase[lSkyTex]] );
		    fSpeedScale = dRealTime * (8.0/1000.0);
		    fSpeedScale -= ((int)fSpeedScale) & ~127;
	
		    for (i = 0, usLastSurf = 0xffff; i < nVS; ++i)
		    {
				int		usSurf = pVS[i];
				texgen_t	*pTG;
				PPRIMDESC	pPrim;
				unsigned int		uCount;

				if (usSurf == usLastSurf) continue;
				usLastSurf = usSurf;

				pTG = &pCurrEpsd->pTG[pS[usSurf].texgenid];
				if (lSkyTex != pTG->texid) continue;

				uCount = pCurrEpsd->pSDL[usSurf].uCount;
				pPrim = &pCurrEpsd->pSPA[pCurrEpsd->pSDL[usSurf].uOffset];

				do {
					short int sMode = pPrim->p.sMode;
				    short int sCount = (pPrim++)->p.sCount;
				    glBegin( sMode );
				    do {
						float s, t, len;
						vec3_t d;

						d[0] = pPrim->c.v[0] - pView->pObject->Head[0];
						d[1] = pPrim->c.v[1] - pView->pObject->Head[1];
						d[2] = pPrim->c.v[2] - pView->pObject->Head[2];
						d[2] *= 3; // flatten the sphere

						len = (6 * 63) / sqrt( d[0] * d[0] + d[1] * d[1] + d[2] * d[2] );

						pPrim->c.lm[0][0] = d[0] * len;
						pPrim->c.lm[0][1] = d[1] * len;

						s = (fSpeedScale + pPrim->c.lm[0][0]) * (1.0 / 128);
						t = (fSpeedScale + pPrim->c.lm[0][1]) * (1.0 / 128);

						glTexCoord2f( s, t );
						glVertex3fv( pPrim->c.v );
						++pPrim;
				    } while (--sCount);
				    glEnd();
				} while (--uCount) ;
		    }

		    // do foreground sky texture
		    glBindTexture( GL_TEXTURE_2D, pCurrEpsd->puTexObj[pCurrEpsd->puTexObjBase[lSkyTex] + 1] );
		    fSpeedScale = dRealTime * (16.0/1000.0);
		    fSpeedScale -= ((int)fSpeedScale) & ~127;
		    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		    glEnable( GL_BLEND );
		    glDepthMask( FALSE );
		    glDepthFunc( GL_LEQUAL );
		    glAlphaFunc( GL_GREATER, 0.1f );
		    glEnable( GL_ALPHA_TEST );

		    for (i = 0, usLastSurf = 0xffff; i < nVS; ++i)
			{
				int		usSurf = pVS[i];
				texgen_t	*pTG;
				PPRIMDESC	pPrim;
				unsigned int		uCount;

				if (usSurf == usLastSurf) continue;
				usLastSurf = usSurf;

				pTG = &pCurrEpsd->pTG[pS[usSurf].texgenid];
				if (lSkyTex != pTG->texid) continue;

				uCount = pCurrEpsd->pSDL[usSurf].uCount;
				pPrim = &pCurrEpsd->pSPA[pCurrEpsd->pSDL[usSurf].uOffset];

				do {
				    short int sMode = pPrim->p.sMode;
				    short int sCount = (pPrim++)->p.sCount;
				    glBegin( sMode );
				    do {
						float s, t;
						s = (fSpeedScale + pPrim->c.lm[0][0]) * (1.0 / 128);
						t = (fSpeedScale + pPrim->c.lm[0][1]) * (1.0 / 128);
						glTexCoord2f( s, t );
						glVertex3fv( pPrim->c.v );
						++pPrim;
					} while (--sCount);
				    glEnd();
				} while (--uCount) ;
		    }

		    glDisable( GL_ALPHA_TEST );
		    glDisable( GL_BLEND );
		    glDepthMask( TRUE );
		    glDepthFunc( GL_LESS );
		}
	    glDisable( GL_TEXTURE_2D );
	}
    else // draw non-textured
    {
		glDisable( GL_TEXTURE_2D );

		for (i = 0, usLastSurf = 0xffff; i < nVS; ++i)
		{
		    int		    usSurf = pVS[i];

			if (usSurf == usLastSurf) continue;
			usLastSurf = usSurf;

			glColor4ub( (GLubyte)((usSurf) & 0xff)
						, (GLubyte)(((usSurf) * (usSurf)) & 0xff)
						, (GLubyte)(0xff - (((usSurf) + 0x80) & 0xff))
						, 0xff );
			DrawSurface( usSurf );
		}
    }
}

void DrawLeafVisibleSet( bspleaf_t *pLeaf, PVIEW pView )
{
	unsigned int	*pVS = (unsigned int *)pCurrEpsd->pTmpLS;
	int			    i, nVS = 0;
	surface_t	    *pS = pCurrEpsd->pS;
	int			    nS = pCurrEpsd->nS;
	int			    bSort;
	#if defined(CULL_STATS)
	int			    nTotal = 0;
	int			    nCulled = 0;
	#endif

	if (pLeaf && (pLeaf->vislist >= 0) && (pLeaf->vislist < pCurrEpsd->nVL))
	{
		unsigned char	*pVL = &pCurrEpsd->pVL[pLeaf->vislist];
		unsigned char	v = *pVL++;
		int				nL = pCurrEpsd->pH[usHull].visleaves;

		bSort = TRUE;
		for (i = 1; i < nL; v = *pVL++)
		{
			if (v)
			{
				int j = 8;
				while (j--)
				{
					// is leaf i visible?
					if (v & 1)
					{
						bspleaf_t *pVisLeaf = &pCurrEpsd->pL[i];
						if (!QuickRejectLeaf( pVisLeaf, pView ))
						{
							int nFirstS = pVisLeaf->firstsurf;
							int nLastS = nFirstS + pVisLeaf->numsurf;
							int j;
							for (j = nFirstS; j < nLastS; ++j)
							{
								pVS[nVS++] = pCurrEpsd->pLS[j];
							}
						}
						#if defined(CULL_STATS)
						else
							++nCulled;
						++nTotal;
						#endif
					}
					++i;
					v >>= 1;
				}
			}
			else i += ((unsigned int)*pVL++) << 3; // skip some leaves
		}
	}
	else // add every surface in the hull
	{
		int	    nFirst = pCurrEpsd->pH[usHull].firstsurf;
		int	    nCnt = pCurrEpsd->pH[usHull].numsurf;

		bSort = FALSE;
		for (i = 0; i < nCnt; ++i)
		{
			pVS[nVS++] = pCurrEpsd->pSortedS[nFirst + i];
		}
	}

	if (bSort)
	{
		// sort surfaces by texture id, then by texgen
		qsort( &pVS[0], nVS, sizeof(pVS[0]), CmpSurfTex );
    }

	if (nVS > pCurrEpsd->nLS)
	{
		printf( "Overflowed temp vis storage (%d > %d)", nVS, pCurrEpsd->nLS );
	}

	#if defined(CULL_STATS)
	if (nTotal > 0) Notify(( "culled %d/%d (%d%%)             \r", nCulled, nTotal, 100 * nCulled / nTotal ));
	#endif

	DrawSurfaceList( pVS, nVS, pView );
}

bspleaf_t *FindBSPLeaf( float x, float y, float z )
{
	bspnode_t		*pNode;
	bspleaf_t		*pLeaf = NULL;

	if (!pCurrEpsd) return( NULL );
	if ((pCurrEpsd->pH[usHull].headnode[0] < 0) || (pCurrEpsd->pH[usHull].headnode[0] >= pCurrEpsd->nN))
		BAIL(( "hull[0] root node out of range!" ));

	pNode = &pCurrEpsd->pN[pCurrEpsd->pH[usHull].headnode[0]];
	while (!pLeaf)
	{
		plane_t *pPlane;
		scalar_t fDist;
		short sNextNode;

		if ((pNode->planenum < 0) || (pNode->planenum >= pCurrEpsd->nP))
		    BAIL(( "plane %d out of range in node!" ));

		pPlane = &pCurrEpsd->pP[pNode->planenum];
		fDist = pPlane->normal[0] * x + pPlane->normal[1] * y + pPlane->normal[2] * z;

		if (fDist > pPlane->dist)
			sNextNode = pNode->front;
		else sNextNode = pNode->back;
		if (sNextNode >= 0)
		{
		    if (sNextNode >= pCurrEpsd->nN)
				BAIL(( "node %d out of range!", sNextNode ));
		    pNode = &pCurrEpsd->pN[sNextNode];
		}
		else // found a leaf
		{
		    sNextNode = ~sNextNode;
		    if (sNextNode >= pCurrEpsd->nL)
				BAIL(( "leaf %d out of range!", sNextNode ));
		    pLeaf = &pCurrEpsd->pL[sNextNode];
		}
    }

Abort:
    return( pLeaf );
}

int IsPointOutsideHull( int nNode, vec3_t *p )
{
    float		x = (*p)[0], y = (*p)[1], z = (*p)[2];

    while (nNode >= 0)
    {
	bsphullnode_t *pNode = &pCurrEpsd->pHN[nNode];
	plane_t *pPlane = &pCurrEpsd->pP[pNode->planenum];
	float fDist;

	fDist = pPlane->normal[0] * x + pPlane->normal[1] * y + pPlane->normal[2] * z - pPlane->dist;

	if (fDist > 0) nNode = pNode->front;
	else nNode = pNode->back;
    }

    return( nNode == CLIP_OUTSIDE );
}

void SetViewport( int bPerspective, int x, int y, int w, int h )
{
    
	glViewport( x, y, w, h );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    if (bPerspective)
    {
		if (w > h)
		{
			double dRatio = h / (double)w;
			glFrustum( -1.0, 1.0, -dRatio, dRatio, 1.0, 5000 );
		}
		else
		{
			double dRatio = w / (double)h;
			glFrustum( -dRatio, dRatio, -1.0, 1.0, 1.0, 5000 );
		}
	}
	else
	{
		glOrtho( -0.5, w - 0.5, -0.5, h - 0.5, -5, 5 );
	}
}

void RenderLevel( void )
{
	bspleaf_t	*pLeaf;
	PVIEW		pView;
	int			i = 0;

	do{
		pView = pCurrEpsd->pView[i];

		pView->View[0] = cos((pView->pObject->fHeading
							+ pView->HAngle) * M_PI / 180.0 );
		pView->View[1] = sin((pView->pObject->fHeading
							+ pView->HAngle) * M_PI / 180.0 );
		pView->View[2] = pView->pObject->View[2];

		SetViewport( TRUE, pView->ScrX, pView->ScrY,
						pView->Width, pView->Height );

		glClear( GL_DEPTH_BUFFER_BIT );
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		gluLookAt( pView->pObject->Head[0],
					pView->pObject->Head[1],
					pView->pObject->Head[2],
					pView->pObject->Head[0] + pView->View[0],
					pView->pObject->Head[1] + pView->View[1],
					pView->pObject->Head[2] + pView->View[2],
					0, 0, 1 );

		pLeaf = FindBSPLeaf( pView->pObject->Head[0],
								pView->pObject->Head[1],
								pView->pObject->Head[2] );

		if (pLeaf)
	    {
			DrawLeafVisibleSet( pLeaf, pView );
			if (bDrawHulls && (usHull == 0))
			{
			    while (++usHull < pCurrEpsd->nH) DrawLeafVisibleSet( NULL, pView );
			    usHull = 0;
			}
	    }
	}
	while(pCurrEpsd->pView[++i] != NULL);

	// set viewport to full window size
	SetViewport( TRUE, 0, 0, W, H );

	BeginOverlay();
	// put all overlays and messages here
		DrawMessage( 10, 5, "Running on %s OpenGL by %s", pGLRenderer, pGLVendor );
		DrawMessage( 10, 20, "%s   fps: %4.2f", EGLE_VERSION, fps );
	EndOverlay();
	glutSwapBuffers();

}

int BuildTextures( void )
{
	int				nNumTex, nAnimTex, nSkyTex, nTotTex, i, nTex;
	int				bSuccess = FALSE;
	unsigned int	*pulRaster = NULL;
	unsigned int	*pulRasterTmp = NULL;
	unsigned int	ulRasterSize = 0;

	if (!pCurrEpsd) goto Abort;
	nNumTex = pCurrEpsd->pMipHead->numtex;
	for (i = nAnimTex = nSkyTex = 0; i < nNumTex; ++i)
	{
		if (pCurrEpsd->pMipTex[i].name[0] == '*')
		{
			ProgressItem(( "found animated texture %s", pCurrEpsd->pMipTex[i].name ));
			++nAnimTex;
		}
		else if (strncmp( pCurrEpsd->pMipTex[i].name, "sky", 3 ) == 0)
		{
			ProgressItem(( "found sky texture %s", pCurrEpsd->pMipTex[i].name ));
			++nSkyTex;
		}
	}
	ProgressItem(( "%d animated textures, %d sky textures", nAnimTex, nSkyTex ));

	nTotTex = nNumTex + nAnimTex * (ANIM_TEX_FRAMES - 1) + nSkyTex * 2;
	pCurrEpsd->puTexObj = malloc( sizeof(pCurrEpsd->puTexObj[0]) * nTotTex );
	if (!pCurrEpsd->puTexObj)
		BAIL(( "Out of memory!" ));
	pCurrEpsd->puTexObjBase = malloc( sizeof(pCurrEpsd->puTexObjBase[0]) * nNumTex );
	if (!pCurrEpsd->puTexObjBase)
		BAIL(( "Out of memory!" ));

	glGenTextures( nTotTex, pCurrEpsd->puTexObj );

	for (i = nTex = 0; i < nNumTex; ++i)
	{
		miptex_t		*pMipTex = &pCurrEpsd->pMipTex[i];
		int				w = pMipTex->width, h = pMipTex->height;
		int				x, y;
		unsigned int	ulSize = w * h;
		unsigned char	*pucTex = pCurrEpsd->ppucTexture[i];

		pCurrEpsd->puTexObjBase[i] = nTex;

		// NULL textures exist in e1m2 and elsewhere...
		if (!pMipTex->name[0]) continue;

		if (ulSize >  ulRasterSize)
		{
			if (pulRaster) free( pulRaster ), pulRaster = NULL;
			// allocate 4x too much
			pulRaster = malloc( 4 * ulSize * sizeof(pulRaster[0]) );
			if (!pulRaster)
				BAIL(( "Out of memory!" ));
			ulRasterSize = 2 * ulSize;
			// setup a scratch raster area too
			pulRasterTmp = &pulRaster[ulRasterSize];
			ProgressItem(( "realloced pulRaster to %d\n", ulRasterSize ));
		}

		if (pMipTex->name[0] == '*') // animated texture
		{
			int			j;
			float		fxScale = 2 * M_PI / w;
			float		fyScale = 2 * M_PI / h;

			for (j = 0; j < ANIM_TEX_FRAMES; ++j)
			{
				float		    fPhase = j * 2 * M_PI / ANIM_TEX_FRAMES;
		
				// build an RGBA raster
				for (x = 0; x < w; ++x)
					for (y = 0; y < h; ++y)
						pulRaster[x + y * w] = Palette[pucTex[x + y * w]];
    
				// shear in x
				for (y = 0; y < h; ++y)
				{
					float	fScramble = (w * ANIM_TEX_DISP) 
										* cos( M_PI_2 * (sin( fPhase )
										+ cos( y * fyScale )) );
					float	fFloor = floor( fScramble );
					float	f2 = fScramble - fFloor;
					float	f1 = 1 - f2;
					int		sx = ((int)fFloor) % w;

					if (sx < 0) sx += w;
					for (x = 0; x < w; ++x)
					{
						unsigned int	ul1 = pulRaster[(x + sx) % w + y * w];
						unsigned int	ul2 = pulRaster[(x + sx + 1) % w + y * w];

						pulRasterTmp[x + y * w] = 
#ifdef ABGR
							0xff000000 |
							((int)((ul1 & 0x00ff0000) * f1 + (ul2 & 0x00ff0000) * f2) & 0x00ff0000) |
							((int)((ul1 & 0x0000ff00) * f1 + (ul2 & 0x0000ff00) * f2) & 0x0000ff00) |
							((int)((ul1 & 0x000000ff) * f1 + (ul2 & 0x000000ff) * f2) & 0x000000ff);
#else
							((int)((ul1 & 0xff000000) * f1 + (ul2 & 0xff000000) * f2) & 0xff000000) |
							((int)((ul1 & 0x00ff0000) * f1 + (ul2 & 0x00ff0000) * f2) & 0x00ff0000) |
							((int)((ul1 & 0x0000ff00) * f1 + (ul2 & 0x0000ff00) * f2) & 0x0000ff00) |
							0xff;
#endif
					}
				}

				// shear in y
				for (x = 0; x < w; ++x)
				{
					float	fScramble = (h * ANIM_TEX_DISP)
										* cos( M_PI_2 * (cos( fPhase )
										+ sin( x * fxScale )) );
					float	fFloor = floor( fScramble );
					float	f2 = fScramble - fFloor;
					float	f1 = 1 - f2;
					int		sy = ((int)fFloor) % h;

					if (sy < 0) sy += h;
					for (y = 0; y < h; ++y)
					{
						unsigned int	ul1 = pulRasterTmp[x + ((y + sy) % h) * w];
						unsigned int	ul2 = pulRasterTmp[x + ((y + sy + 1) % h) * w];

						pulRaster[x + y * w] =
#ifdef ABGR
							0xff000000 |
							((int)((ul1 & 0x00ff0000) * f1 + (ul2 & 0x00ff0000) * f2) & 0x00ff0000) |
							((int)((ul1 & 0x0000ff00) * f1 + (ul2 & 0x0000ff00) * f2) & 0x0000ff00) |
							((int)((ul1 & 0x000000ff) * f1 + (ul2 & 0x000000ff) * f2) & 0x000000ff);
#else
							((int)((ul1 & 0xff000000) * f1 + (ul2 & 0xff000000) * f2) & 0xff000000) |
							((int)((ul1 & 0x00ff0000) * f1 + (ul2 & 0x00ff0000) * f2) & 0x00ff0000) |
							((int)((ul1 & 0x0000ff00) * f1 + (ul2 & 0x0000ff00) * f2) & 0x0000ff00) |
							0xff;
#endif
					}
				}
				glBindTexture( GL_TEXTURE_2D, pCurrEpsd->puTexObj[nTex++] );
				gluBuild2DMipmaps( GL_TEXTURE_2D, nTexelFormat, w, h, nPixelFormat, GL_UNSIGNED_BYTE, pulRaster );
				ProgressItem(( "built anim texture %s frame %d as texid[%d] -> texobj[%d] -> %d\n", pMipTex->name, j, i, nTex - 1, pCurrEpsd->puTexObj[nTex - 1] ));
			}
		}
		else if ((pMipTex->name[0] == 's') && 
				(pMipTex->name[1] == 'k') && 
				(pMipTex->name[2] == 'y')) // sky texture
		{
			int				w2 = w / 2; // sky is actually two textures side-by-side
			int				ar, ag, ab;
			int				mr, mg, mb, nMax;
			unsigned int	ulAvg;

			// build an RGBA raster for the background level, find average pixel value
			ar = ag = ab = 0;
			mr = mg = mb = 0;
			for (x = 0; x < w2; ++x)
				for (y = 0; y < h; ++y)
				{
					unsigned int	ulRGBA = Palette[pucTex[x + w2 + y * w]];
#ifdef ABGR
					int				r = (ulRGBA & 0xff);
					int				g = ((ulRGBA >>  8) & 0xff);
					int				b = ((ulRGBA >> 16) & 0xff);
#else
					int				r = ((ulRGBA >> 24) & 0xff);
					int				g = ((ulRGBA >> 16) & 0xff);
					int				b = ((ulRGBA >>  8) & 0xff);
#endif
					pulRaster[x + y * w2] = ulRGBA;
					ar += r; ag += g; ab += b;
					if (mr < r)
						mr = r;
					if (mg < g)
						mg = g;
					if (mb < b)
						mb = b;
				}
			glBindTexture( GL_TEXTURE_2D, pCurrEpsd->puTexObj[nTex++] );
			gluBuild2DMipmaps( GL_TEXTURE_2D, nTexelFormat, w2, h, nPixelFormat, GL_UNSIGNED_BYTE, pulRaster );
			ProgressItem(( "built bg sky texture %s as texid[%d] -> texobj[%d] -> %d\n", pMipTex->name, i, nTex - 1, pCurrEpsd->puTexObj[nTex - 1] ));

#ifdef ABGR
			ulAvg = ((ar / (w2 * h)) & 0xff) |
					((ag / (w2 * h)) & 0xff) << 8 | 
					((ab / (w2 * h)) & 0xff) << 16;
#else
			ulAvg = ((ar / (w2 * h)) & 0xff) << 24 |
					((ag / (w2 * h)) & 0xff) << 16 | 
					((ab / (w2 * h)) & 0xff) << 8;
#endif
			nMax = mr + mg + mb + 1;

			// build an RGBA raster for the foreground level
			for (x = 0; x < w2; ++x)
				for (y = 0; y < h; ++y)
				{
					unsigned char	uc = pucTex[x + y * w];
					if (uc)
					{
						unsigned int	ulColor = Palette[uc];
#ifdef ABGR
						unsigned int	ulAlpha = (((ulColor & 0xff)
										+ ((ulColor >> 8) & 0xff) 
										+ ((ulColor >> 16) & 0xff)) * 180) / nMax;
						ulColor = (ulColor & ~0xff000000) | ((ulAlpha << 24) & 0xff000000);
#else
						unsigned int	ulAlpha = ((((ulColor >> 24) & 0xff)
										+ ((ulColor >> 16) & 0xff) 
										+ ((ulColor >> 8) & 0xff)) * 180) / nMax;
						ulColor = (ulColor & ~0xff) | (ulAlpha & 0xff);
#endif
						pulRaster[x + y * w2] = ulColor;
					}
					else
						pulRaster[x + y * w2] = ulAvg;
				}
			glBindTexture( GL_TEXTURE_2D, pCurrEpsd->puTexObj[nTex++] );
			gluBuild2DMipmaps( GL_TEXTURE_2D, nTexelFormat, w2, h, nPixelFormat, GL_UNSIGNED_BYTE, pulRaster );
			ProgressItem(( "built fg sky texture %s as texid[%d] -> texobj[%d] -> %d\n", pMipTex->name, i, nTex - 1, pCurrEpsd->puTexObj[nTex - 1] ));
		}
	
		else if (pMipTex->name[0] == '#') // RGB texture
		{
			// build an RGBA raster
			for (x = 0; x < w * h; ++x)
				pulRaster[x] = ((((unsigned int)pucTex[x * 3]))
							| (((unsigned int)pucTex[x * 3 + 1]) << 8)
							| (((unsigned int)pucTex[x * 3 + 2]) << 16)
							| (( 0xff ) << 24));    
			glBindTexture( GL_TEXTURE_2D, pCurrEpsd->puTexObj[nTex++] );
			gluBuild2DMipmaps( GL_TEXTURE_2D, nTexelFormat, w, h, nPixelFormat, GL_UNSIGNED_BYTE, pulRaster );
			ProgressItem(( "built texture %s as texid[%d] -> texobj[%d] -> %d\n", pMipTex->name, i, nTex - 1, pCurrEpsd->puTexObj[nTex - 1] ));
		}
		else // normal texture
		{
			// build an RGBA raster
			for (x = 0; x < w; ++x)
				for (y = 0; y < h; ++y)
					pulRaster[x + y * w] = Palette[pucTex[x + y * w]];
			glBindTexture( GL_TEXTURE_2D, pCurrEpsd->puTexObj[nTex++] );
			gluBuild2DMipmaps( GL_TEXTURE_2D, nTexelFormat, w, h, nPixelFormat, GL_UNSIGNED_BYTE, pulRaster );
			ProgressItem(( "built texture %s as texid[%d] -> texobj[%d] -> %d\n", pMipTex->name, i, nTex - 1, pCurrEpsd->puTexObj[nTex - 1] ));
		}
	}

	bSuccess = TRUE;

Abort:
	if (pulRaster) free( pulRaster ), pulRaster = NULL;
	return( bSuccess );
}

int CalcTexCoords( void )
{
    surface_t	*pS = pCurrEpsd->pS;
    int			i;

    // texgen is slow on some machines, precalculate texcoords here
    for (i = 0; i < pCurrEpsd->nS; ++i)
    {
	int		    k;
	unsigned int		j;
	texgen_t	    *pTG = &pCurrEpsd->pTG[pS[i].texgenid];
	miptex_t	    *pMipTex = &pCurrEpsd->pMipTex[pTG->texid];
	float		    xScale, yScale;

	// setup the display-listed texture matrix
	xScale = 1.0 / pMipTex->width;
	yScale = 1.0 / pMipTex->height;

	for (j = 0; j < pCurrEpsd->pSDL[i].uCount; ++j)
	{
	    PPRIMDESC p = &pCurrEpsd->pSPA[pCurrEpsd->pSDL[i].uOffset + j];
	    
	    for (k = (p++)->p.sCount; k > 0; --k, ++p)
	    {
		// find texture coords
		p->c.t[0] = (DotVec3( pTG->snrm, p->c.v ) + pTG->soff) * xScale;
		p->c.t[1] = (DotVec3( pTG->tnrm, p->c.v ) + pTG->toff) * yScale;

	    }
	}
    }

    return( TRUE );
}

